<?php

namespace smtech\LTI\Configuration;

use DOMDocument;
use DOMElement;

use smtech\LTI\Configuration\LaunchPrivacy;
use smtech\LTI\Configuration\Option;
use smtech\LTI\Configuration\Exception\ConfigurationException;

/**
 * Generate a valid LTI Tool Provider configuration XML document to facilitate
 * tool placements within a Tool Consumer.
 *
 * With minimal information (a name, ID and launch URL), the default
 * configuration will provide a course navigation placement that uses the name
 * as the link text and the launch URL as the the link URL, defaulting to
 * 'anonymous' launch privacy.
 *
 * @see https://www.edu-apps.org/build_xml.html Edu Apps XML Config Builder
 *      provides an interactive interface for generating static configuration
 *      XML files.
 *
 * @author Seth Battis <SethBattis@stmarksschool.org>
 * @version v1.0
 */
class Generator
{

    /** XML schema instance */
    const XSI =   'http://www.w3.org/2001/XMLSchema-instance';

    /** XML namespacing */
    const XMLNS = 'http://www.w3.org/2000/xmlns/';

    /**
     * This is the description of the resource linkfile that is to be placed
     * inside a Common Cartridge.
     */
    const LTICC = 'http://www.imsglobal.org/xsd/imslticc_v1p0';

    /**
     * The BasicLTILink class is the container for information required to use
     * the BasicLTI mechanism.
     */
    const BLTI =  'http://www.imsglobal.org/xsd/imsbasiclti_v1p0';

    /**
     * This is the description of the Common Messaging objects in LTI.
     */
    const LTICM = 'http://www.imsglobal.org/xsd/imslticm_v1p0';

    /**
     * This is the set of Common Profile objects used in LTI.
     */
    const LTICP = 'http://www.imsglobal.org/xsd/imslticp_v1p0';

    /**
     * Human-readable name of the Tool Provider
     *
     * @var string
     */
    protected $name;

    /**
     * Globally-unique tokenized ID for the Tool Provider
     *
     * @var string
     */
    protected $id;

    /**
     * Launch URL to which LTI request will be passed by Tool Consumer
     *
     * Tool placement options that do not specify a launch URL will inherit
     * this URL.
     *
     * @var string
     */
    protected $launchURL;

    /**
     * Human-readable description of the Tool Provider
     *
     * This setting will be ignored if it is not set.
     *
     * @var string
     */
    protected $description = false;

    /**
     * URL of the 16x16 pixel icon for the Tool provider
     *
     * This setting will be ignored if it is not set.
     *
     * @var string
     */
    protected $iconURL = false;

    /**
     * Custom parameters to be passed to the Tool Provider in the initial
     * launch request
     *
     * This setting will be ignored if it is not set.
     *
     * @var string[] An associative array of keys and values
     */
    protected $custom = [];

    /**
     * The level of user information privacy requested of the Tool Consumer
     *
     * This setting will default to anonymous if not set.
     *
     * @var string Must be a valid value of `LaunchPrivacy` enumerated type.
     *
     * @see LaunchPrivacy `LaunchPrivacy` enumerated type
     */
    protected $launchPrivacy = false;

    /**
     * Expected domain from which the Tool Consumer will make requests
     *
     * This setting will be ignored if it is not set
     *
     * @var string
     */
    protected $domain = false;

    /**
     * Tool placement options configurations
     *
     * This setting will default to a course navigation placement if it is not
     * set.
     *
     * @var array Associative array of with valid values of `Option`
     *      enumerated type as keys to associative arrays of name/value pairs
     *      of parameters for that option: `[Option => [$name => $value]]`
     *
     * @see Option `Option` enumerated type
     */
    protected $options = [];

    /**
     * Construct a Generator
     *
     * @param string $name
     * @param string $id
     * @param string $launchURL
     * @param string $description (Optional)
     * @param string $iconURL (Optional)
     * @param string $launchPrivacy (Optional, defaults to anonymous) Must be
     *     a valid value of the `LauchPrivacy` enumerated type
     * @param string $domain (Optional)
     */
    public function __construct(
        $name,
        $id,
        $launchURL,
        $description = null,
        $iconURL = null,
        $launchPrivacy = null,
        $domain = null
    ) {
        $this->setName($name);
        $this->setID($id);
        $this->setLaunchURL($launchURL);
        $this->setDescription($description);
        $this->setIconURL($iconURL);
        $this->setLaunchPrivacy($launchPrivacy);
        $this->setDomain($domain);
    }

    /**
     * Set human-readable name of Tool Provider
     *
     * @param string $name
     */
    public function setName($name)
    {
        if (empty((string) $name)) {
            throw new ConfigurationException(
                'The configuration must specify a non-empty name for the Tool Provider.',
                ConfigurationException::TOOL_PROVIDER
            );
        } else {
            $this->name = (string) $name;
        }
    }

    /**
     * Set globally unique, tokenized ID for Tool Provider
     *
     * @param string $id [description]
     * @throws ConfigurationException TOOL_PROVIDER if empty
     */
    public function setID($id)
    {
        /*
         * TODO validate/tokenize the actual ID
         */
        if (empty((string) $id)) {
            throw new ConfigurationException(
                'The configuration must specify a non-empty (and globally unique) ID for the Tool Provider.',
                ConfigurationException::TOOL_PROVIDER
            );
        } else {
            $this->id = (string) $id;
        }
    }

    /**
     * Set the launch URL to which the Tool Consumer will pass the initial
     * request
     *
     * @param string $launchURL
     */
    public function setLaunchURL($launchURL)
    {
        /*
         * TODO validate the actual URL?
         */
        if (empty((string) $launchURL)) {
            throw new ConfigurationException(
                'The configuration must specify a valid launch URL for the Tool Provider.',
                ConfigurationException::TOOL_PROVIDER
            );
        } else {
            $this->launchURL = (string) $launchURL;
        }
    }

    /**
     * Set the human-readable description of the Tool Provider
     *
     * @param string $description
     */
    public function setDescription($description)
    {
        $this->description = (empty((string) $description) ? false : (string) $description);
    }

    /**
     * Set URL of 16x16 pixel icon image for Tool Provider
     *
     * @param string $iconURL
     */
    public function setIconURL($iconURL)
    {
        /*
         * TODO validate the actual URL?
         */
        $this->iconURL = (empty((string) $iconURL) ? false : (string) $iconURL);
    }

    /**
     * Set the level of user profile privacy requested of the Tool Consumer
     *
     * @param string $launchPrivacy Must be a valid value of the
     *     `LaunchPrivacy` enumerated type
     *
     * @see LaunchPrivacy `LaunchPrivacy` enumerated type
     */
    public function setLaunchPrivacy($launchPrivacy)
    {
        if (!is_a($launchPrivacy, LaunchPrivacy::class) && !LaunchPrivacy::isValid($launchPrivacy)) {
            throw new ConfigurationException(
                "Invalid launch privacy setting '$launchPrivacy'",
                ConfigurationException::TOOL_PROVIDER
            );
        } else {
            $this->launchPrivacy = (empty($launchPrivacy) ? LaunchPrivacy::ANONYMOUS() : $launchPrivacy);
        }
    }

    /**
     * Set the domain from which Tool Consumer requests will come
     *
     * @param string $domain
     */
    public function setDomain($domain)
    {
        $this->domain = (empty($domain) ? false : $domain);
    }

    /**
     * Set a tool placement option
     *
     * @param string $option Must be a valid value of the `Option` enumerated
     *     type
     * @param string[]  $properties Associative array of option properties in
     *     name/value pairs
     *
     * @see Option `Option` enumerated type
     */
    public function setOption($option, array $properties)
    {
        if (!is_a($option, Option::class) && !Option::isValid($option)) {
            throw new ConfigurationException(
                "Invalid configuration option '$option'",
                ConfigurationException::TOOL_PROVIDER
            );
        } else {
            $this->options[(string) $option] = $properties;
        }
    }

    /**
     * Set a particular property of a particular tool placment option
     *
     * @param string $option Must be a valid value of the `Option` enumerated
     *     type
     * @param string $property
     * @param string $value
     *
     * @see Option `Option` enumerated type
     */
    public function setOptionProperty($option, $property, $value)
    {
        if (!is_a($option, Option::class) && !Option::isValid($option)) {
            throw new ConfigurationException(
                "Invalid configuration option '$option'",
                ConfigurationException::TOOL_PROVIDER
            );
        } else {
            $this->options[(string) $option][$property] = $value;
        }
    }

    /**
     * Generate the XML configuration file
     *
     * @return string
     */
    public function saveXML()
    {

        $config = new DOMDocument('1.0', 'UTF-8');
        $config->formatOutput = true;

        $cartridge = $config->createElementNS(self::LTICC, 'cartridge_basiclti_link');
        $config->appendChild($cartridge);
        $cartridge->setAttributeNS(self::XMLNS, 'xmlns:blti', self::BLTI);
        $cartridge->setAttributeNS(self::XMLNS, 'xmlns:lticm', self::LTICM);
        $cartridge->setAttributeNS(self::XMLNS, 'xmlns:lticp', self::LTICP);
        $cartridge->setAttributeNS(self::XMLNS, 'xmlns:xsi', self::XSI);
        $cartridge->setAttributeNS(
            self::XSI,
            'xsi:schemaLocation',
            self::LTICC . ' ' . self::LTICC . '.xsd ' .
            self::BLTI . ' ' . self::BLTI . '.xsd ' .
            self::LTICM . ' ' . self::LTICM . '.xsd ' .
            self::LTICP . ' ' . self::LTICP . '.xsd'
        );

        $cartridge->appendChild($config->createElementNS(
            self::BLTI,
            'blti:title',
            $this->name
        ));

        /*
         * TODO CDATA wrapper?
         */
        if ($this->description) {
            $cartridge->appendChild($config->createElementNS(
                self::BLTI,
                'blti:description',
                $this->description
            ));
        }

        if ($this->iconURL) {
            $cartridge->appendChild($config->createElementNS(
                self::BLTI,
                'blti:icon',
                $this->iconURL
            ));
        }

        $cartridge->appendChild($config->createElementNS(
            self::BLTI,
            'blti:launch_url',
            $this->launchURL
        ));

        $extensions = $config->createElementNS(self::BLTI, 'blti:extensions');
        $cartridge->appendChild($extensions);
        $extensions->setAttribute('platform', 'canvas.instructure.com');

        $property = $config->createElementNS(
            self::LTICM,
            'lticm:property',
            $this->id
        );
        $property->setAttribute('name', 'tool_id');
        $extensions->appendChild($property);

        $property = $config->createElementNS(
            self::LTICM,
            'lticm:property',
            $this->launchPrivacy
        );
        $property->setAttribute('name', 'privacy_level');
        $extensions->appendChild($property);

        if (!empty($this->domain)) {
            $property = $config->createElementNS(
                self::LTICM,
                'lticm:property',
                $this->domain
            );
            $property->setAttribute('name', 'domain');
            $extensions->appendChild($property);
        }

        /* if no options are configured, create a default course navigation option */
        if (empty($this->options)) {
            $extensions->appendChild($this->getOptionsElement(
                $config,
                Option::COURSE_NAVIGATION(),
                []
            ));
        } else {
            foreach ($this->options as $option => $properties) {
                $extensions->appendChild($this->getOptionsElement(
                    $config,
                    $option,
                    $properties
                ));
            }
        }

        $bundle = $config->createElement('cartridge_bundle');
        $cartridge->appendChild($bundle);
        $bundle->setAttribute('identiferref', 'BLT001_Bundle');

        $icon = $config->createElement('cartridge_icon');
        $cartridge->appendChild($icon);
        $icon->setAttribute('identifierref', 'BLT001_Icon');

        return $config->saveXML();
    }

    /**
     * Build a tool placement option element
     *
     * @param  DOMDocument $config
     * @param  Option $option
     * @param  string[] $properties
     * @return DOMElement
     */
    private function getOptionsElement(DOMDocument $config, $option, array $properties)
    {
        if (is_a($option, Option::class) || Option::isValid($option)) {
            $options = $config->createElementNS(self::LTICM, 'lticm:options');
            $options->setAttribute('name', $option);

            /* inherit link text and launch URL properties if not specified */
            if (!array_key_exists('text', $properties)) {
                $properties['text'] = $this->name;
            }
            if (!array_key_exists('url', $properties)) {
                $properties['url'] = $this->launchURL;
            }

            foreach ($properties as $name => $value) {
                $property = $config->createElementNS(
                    self::LTICM,
                    'lticm:property',
                    $value
                );
                $property->setAttribute('name', $name);
                $options->appendChild($property);
            }

            return $options;
        }
        return null;
    }
}
